/* -*- mode: c++; c-basic-offset: 3; -*- */
#include "YSynch.hh"
#include <stdexcept>
#include "DVector.hh"

using namespace std;

//======================================  YSynch constructor
YSynch::YSynch(const YPipe& yp, synch_mode mode)
   : _mode(mode)
{
   _pipe.reset(yp.clone());
}

//======================================  YSynch destructor
YSynch::~YSynch(void) {
}

YSynch* 
YSynch::clone(void) const {
   return new YSynch(*_pipe, _mode);
}

//======================================  Synchronize the data and apply filter.
TSeries 
YSynch::apply(const TSeries& x, const TSeries& y) {
   if (!_pipe) throw logic_error("YSynch: No YPipe specified");
   TSeries ts_out;

   //-----------------------------------  Let the YPipe synch itself
   bool     mtq = _xqueue.empty() || _yqueue.empty();
   bool synchOK = !mtq && (x.getStartTime() == y.getStartTime()) &&
                          (x.getInterval() == y.getInterval());     
   if (_mode == m_except || synchOK) {
      if (!inUse()) {
	 mStartTime   = x.getStartTime();
	 mCurrentTime = x.getEndTime();
      }
      return _pipe->apply(x, y);
   }

   //-----------------------------------  Fill in queue
   if (!inUse()) {
      int fail = _xqueue.Append(x);
      if (fail) throw ("YSynch: input data are not contiguous");
      fail = _yqueue.Append(y);
      if (fail) throw ("YSynch: input data are not contiguous");
   }
   if (_xqueue.empty() || _yqueue.empty()) return ts_out;
   
   //-----------------------------------  Pad starting data
   if (_mode == m_pad) {
      Interval tDiff = _xqueue.getStartTime() - _yqueue.getStartTime();
      if (double(tDiff) > 0.0) {
	 Interval dT = _xqueue.getTStep();
	 size_t nInsert = size_t(tDiff / dT + 0.5);
	 Time   tStartx = _xqueue.getStartTime() - dT * double(nInsert);
	 if (!Almost(tStartx, _yqueue.getStartTime())) {
	    throw runtime_error("YSynch: Unable to synchronize start times");
	 }
	 if (nInsert) {
	    DVector* dv = _xqueue.refDVect()->clone();
	    dv->replace_with_zeros(0, 0, nInsert);
	    _xqueue.setData(tStartx, dT, dv);
	 }
      }
      else if (double(tDiff) < 0.0) {
	 Interval dT = _yqueue.getTStep();
	 size_t nInsert = size_t(-tDiff / dT + 0.5);
	 Time   tStarty = _yqueue.getStartTime() - dT * double(nInsert);
	 if (!Almost(tStarty, _xqueue.getStartTime())) {
	    throw runtime_error("YSynch: Unable to synchronize start times");
	 }
	 if (nInsert) {
	    DVector* dv = _yqueue.refDVect()->clone();
	    dv->replace_with_zeros(0, 0, nInsert);
	    _yqueue.setData(tStarty, dT, dv);
	 }
      }
   }
   else if (_mode == m_trash) {
      Interval tDiff = _xqueue.getStartTime() - _yqueue.getStartTime();
      if (double(tDiff) > 0.0) {
	 _yqueue.eraseStart(tDiff);
      } else {
	 _xqueue.eraseStart(-tDiff);
      }
   }
   else {
      throw  logic_error("YSynch: No synchronization mode specified");
   }
   Interval len = _xqueue.getInterval();
   if (_yqueue.getInterval() < len) len = _yqueue.getInterval();
   Time tStart = _xqueue.getStartTime();
   ts_out = _pipe->apply(_xqueue.extract(tStart, len),
			 _yqueue.extract(tStart, len));
   _xqueue.eraseStart(len);
   _yqueue.eraseStart(len);
   return ts_out;
}
//======================================  Get the time delay from the filter
void 
YSynch::dataCheck(const TSeries& x, const TSeries& y) const {
   bool     mtq = _xqueue.empty() || _yqueue.empty();
   bool synchOK = !mtq && (x.getStartTime() == y.getStartTime()) &&
                          (x.getInterval() == y.getInterval());     
    if (_mode == m_except || synchOK) {
      _pipe->dataCheck(x, y);
   }
}

//======================================  Get the time delay from the filter
Interval
YSynch::getTimeDelay(void) const {
   return _pipe->getTimeDelay();
}

//======================================  Reset the filter and synchronization
//                                        queue states
void
YSynch::reset(void) {
   mStartTime = Time(0);
   mCurrentTime = mStartTime;
   _pipe->reset();
   _xqueue.Clear();
   _yqueue.Clear();
}

